import argparse
import copy
import functools
import itertools
import os
import pandas as pd
import socket
import subprocess
from dataclasses import dataclass
from pathlib import Path
from typing import Dict, List, Union, Optional
from uuid import uuid4
import matplotlib.pyplot as plt
import seaborn as sns
from ruamel.yaml import YAML
import plotly.graph_objects as go
from osin.config import ROOT_DIR
from osin.db import Job, ExpResult


# noinspection PyMethodMayBeStatic
@dataclass
class ExpConfig:
    """Description of an experiment. It includes parameters and their possible values.
    """
    # name of the report
    name: str
    # name of the table to store the run
    table_name: str
    # description of the run
    description: str
    # parameters of the report and their possible values
    parameters: Dict[str, List[str]]
    # the list of reports we should made from the experiments
    reports: List[dict]
    # command to trigger the run to gather data for the report
    run_trigger_type: str
    run_trigger_args: List[str]

    @staticmethod
    def from_file(infile: Union[str, Path]) -> List['ExpConfig']:
        with open(str(infile), "r") as f:
            yaml = YAML()
            data = yaml.load(f)

        reports = []
        for report in data:
            # allow args to be either string or list
            if isinstance(report['run_trigger']['args'], list):
                args = report['run_trigger']['args']
            else:
                args = report['run_trigger']['args'].split(" ")

            report = ExpConfig(
                name=report['name'],
                table_name=report.get('table', 'default'),
                description=report.get('description', ''),
                parameters=report['parameters'],
                reports=report['reports'],
                run_trigger_type=report['run_trigger']['type'],
                run_trigger_args=args
            )
            reports.append(report)
        return reports

    def trigger_runs(self, parameters: Dict[str, List[str]]):
        args = self.run_trigger_args
        runs = []

        # find place to put the parameters in the run trigger arguments
        param_placements = {}
        for i, arg in enumerate(args):
            if arg.startswith("%%") and arg.endswith("%%") and arg[2:-2] in parameters:
                param_placements[arg[2:-2]] = i

        for values in itertools.product(*parameters.values()):
            run_args = copy.copy(args)
            for k, v in zip(parameters.keys(), values):
                run_args[param_placements[k]] = v
            runs.append(run_args)

        jobs = []
        if self.run_trigger_type == "bash":
            for run_args in runs:
                split_idx = run_args.index("--")
                init_args, rargs = run_args[:split_idx], run_args[split_idx + 1:]
                jobs.append(Job.create(
                    exec_type=self.run_trigger_type,
                    exec_init_args=init_args,
                    exec_run_args=rargs,
                    status="queueing"
                ))
        else:
            raise NotImplementedError()
        return jobs

    def report(self, names: List[str] = None):
        if names is None:
            names = [report['name'] for report in self.reports]

        assert len(names) == len(set(names)), "Duplicated report names"
        # get the data first
        df = ExpResult.as_dataframe(self.table_name)

        def get_value_func(name: str):
            report = [report for report in self.reports if report['name'] == name][0]
            fname = f'report_{report["type"]}'
            if not hasattr(self, fname):
                raise NotImplementedError(f"Not support {report['type']}")
            func = getattr(self, fname)

            @functools.wraps(func)
            def func_wrapper():
                # handle no data
                if len(df) == 0:
                    return f"No data for report {report['type']}"

                try:
                    return func(df, **report.get('params', {}))
                except Exception as e:
                    # return an exception for streamlit to display
                    return e
            return func_wrapper

        report_grids = [[]]
        for name in names:
            report = [report for report in self.reports if report['name'] == name][0]
            colspan = report.get('colspan', 24)
            row_used_size = sum([r['colspan'] for r in report_grids[-1]])
            if colspan == 24 or row_used_size + colspan > 24:
                # use new row
                report_grids.append([])
            report_grids[-1].append({
                "name": name,
                "colspan": colspan,
                "display_name": report.get('display_name', True),
                "get_value": get_value_func(name)
            })

        if len(report_grids[0]) == 0:
            report_grids.pop(0)

        return report_grids

    def report_matrix(self, df: pd.DataFrame, columns: List[str], agg_metrics: List[str] = None,
                      groupby: List[str] = None):
        agg_metrics = agg_metrics or ['mean', 'min', 'max', 'std']
        if groupby is not None:
            df = df.groupby(groupby)
        return df[columns].aggregate(agg_metrics)

    def report_barplot(self, df: pd.DataFrame, x: str, y: str, group: str = None, title: str = None):
        if group is not None:
            group_values = df[group].unique()
            sdfs = [(group_value, df[df[group] == group_value]) for group_value in group_values]
        else:
            sdfs = [(None, df)]

        data = []
        for group_value, sdf in sdfs:
            sdf = sdf.groupby(x).aggregate(['mean', 'min', 'max'])
            args = dict(x=sdf.index,
                        y=sdf[y, 'mean'],
                        text=[f"{v:.3f}" for v in sdf[y, 'mean']],
                        textposition='auto',
                        error_y=dict(
                            type='data',
                            symmetric=False,
                            array=sdf[y, 'max'] - sdf[y, 'mean'],
                            arrayminus=sdf[y, 'mean'] - sdf[y, 'min'],
                            visible=True
                        ))
            if group is not None:
                args["name"] = group
            data.append(args)

        fig = go.Figure(data=[
            go.Bar(**kwargs)
            for kwargs in data
        ])

        if title is not None:
            fig.update_layout(title_text=title)
        return fig


# if __name__ == '__main__':
#     reports = ExpConfig.from_file(os.path.join(ROOT_DIR, "experiments.yml"))
#     # reports[0].trigger_runs({k: v for k, v in reports[0].parameters.items()})
#     reports[0].report()
#     # print(list(Job.select().where(Job.hostname == 'sequoia', Job.pid == '12950')))



import argparse
import os
import socket
import subprocess
import time
from uuid import uuid4
from loguru import logger
from osin.db import Job, db


class JobExecutor:
    def __init__(self, check_interval: int = 1):
        # find the last finished job
        self.check_interval = check_interval

        self.last_job_id = Job.last_finished_job()
        self.bash_argparser = argparse.ArgumentParser(description="Run bash command")
        self.bash_argparser.add_argument("--cwd", help="Working directory")

    def start(self):
        """Start a daemon monitoring the database for new jobs to run"""
        logger.info("Worker started!")
        printed_wait_msg = False
        while True:
            # find new jobs
            jobs = list(Job.select().where(Job.id > self.last_job_id))
            if len(jobs) > 0:
                logger.info("Found {} jobs", len(jobs))
                printed_wait_msg = False
            else:
                if not printed_wait_msg:
                    logger.info("Waiting for jobs...")
                    printed_wait_msg = True
            for job in jobs:
                if job.exec_type == "bash":
                    logger.info("Execute job: {}", job.id)
                    tracker = self.bash(job)
                    self.wait_till_finish(tracker)
                    self.last_job_id = job.id

                    with db.atomic():
                        new_job = Job.get_by_id(job.id)
                        assert new_job.status != "queueing"
                        if new_job.status == "started":
                            # this job has failed to report to the server, so we mark it as failed
                            new_job.status = "failure"
                            new_job.save()
                    logger.info("Execute job: {}. Done!", job.id)
                else:
                    raise NotImplementedError(job.exec_type)

            # rest before checking again
            time.sleep(self.check_interval)

    def bash(self, job: Job) -> subprocess.Popen:
        """Run program in bash and save the job to db"""
        bash_args = self.bash_argparser.parse_args(job.exec_init_args)
        logfile = f"/tmp/{str(uuid4())}.log"
        with open(logfile, "w") as f:
            p = subprocess.Popen(job.exec_run_args, cwd=bash_args.cwd,
                                 stdout=f, stderr=f,
                                 start_new_session=True, env=dict(os.environ))
        job.hostname = socket.gethostname()
        job.pid = str(p.pid)
        job.status = "started"
        job.save()
        return p

    def wait_till_finish(self, p: subprocess.Popen):
        p.wait()


if __name__ == '__main__':
    JobExecutor().start()



import os
import socket

import requests


def submit(data: dict, table: str = 'default', host: str = 'http://localhost:5524'):
    """Submit results to the server"""
    resp = requests.post(host + "/api/v1/runs", json={
        "data": data,
        "table": table,
        "hostname": socket.gethostname(),
        "pid": str(os.getpid())
    })
    assert resp.status_code == 200, resp


def error(host: str = 'http://localhost:5524'):
    """Tell the server that we encounter some errors. This is optional as we monitor for changes"""
    resp = requests.post(host + "/api/v1/runs/error", json={
        "hostname": socket.gethostname(),
        "pid": str(os.getpid())
    })
    assert resp.status_code == 200, resp



"""Mockup experiments
"""
import random
import time

import click
from osin.api import submit


@click.command()
@click.option("--dataset", help="dataset")
@click.option("--method", help="method")
@click.option("--exectime", help="execution time in seconds")
def main(dataset, method, exectime):
    precision = random.randint(70, 100) / 100
    recall = random.randint(70, 100) / 100
    f1 = 2 * precision * recall / (precision + recall)
    k = 10
    for i in range(k):
        print("Training batch", i)
        time.sleep(int(exectime) / k)
    print("Training done!")
    submit({
        "dataset": dataset,
        "method": method,
        "precision": precision,
        "recall": recall,
        "f1": f1,
        "exectime": exectime
    })


if __name__ == '__main__':
    main()


from datetime import datetime
from typing import List, Optional

import pandas as pd
from peewee import *
from playhouse.sqlite_ext import JSONField
from operator import attrgetter
from osin.config import DBFILE

db = SqliteDatabase(DBFILE)


class ExpResult(Model):
    class Meta:
        database = db

    table = TextField(default="default")
    created_time = DateTimeField(default=datetime.now)
    data = JSONField()

    @staticmethod
    def as_dataframe(table_name: str="default") -> pd.DataFrame:
        """Convert the whole table into a single data frame"""
        with db.atomic():
            records = [dict(r.data, created_time=r.created_time) for r in ExpResult.select().where(ExpResult.table == table_name)]
            return pd.DataFrame(records)

    @staticmethod
    def merge_column(column_1: str, column_2: str, table_name: str="default"):
        with db.atomic():
            # TODO: do bulk update
            for r in ExpResult.select().where(ExpResult.table == table_name):
                if column_1 not in r.data and column_2 not in r.data:
                    continue
                assert r.data.get(column_1, None) is None or r.data.get(column_2, None) is None, "Can't overwrite existing data"
                if r.data.get(column_1, None) is None:
                    r.data[column_1] = r.data.get(column_2, None)
                if column_2 in r.data:
                    del r.data[column_2]
                r.save()


class Job(Model):
    class Meta:
        database = db

    created_time = DateTimeField(default=datetime.now, index=True)
    start_time = DateTimeField(null=True, default=None)
    exec_type = TextField(choices=["bash"])
    exec_init_args = JSONField()
    exec_run_args = JSONField()
    hostname = TextField(null=True)
    pid = TextField(null=True)
    logfile = TextField(null=True)
    status = TextField(choices=["queueing", "started", "success", "failure"], index=True)

    @staticmethod
    def last_finished_job() -> int:
        jobs = Job.select(Job.id).where(Job.status.in_(['failure', 'success'])).order_by(Job.id.desc()).limit(1)
        jobs = list(jobs)
        if len(jobs) == 0:
            return 0
        else:
            return jobs[0].id

    @staticmethod
    def get_running_jobs() -> List['Job']:
        cursor = Job.select() \
            .where(Job.status.in_(['queueing', 'started'])) \
            .order_by(Job.id.desc())
        return list(cursor)

    def is_queueing(self):
        return self.status == "queueing"

    def is_started(self):
        return self.status == "started"


db.create_tables([ExpResult, Job])


if __name__ == '__main__':
    Job.recent_jobs(0, 10)
#     RunTable.create(data={"method": "random_forest", "precision": 1, "recall": 0.5})
#     RunTable.create(data={"method": "random_forest", "precision": 3, "recall": 0.5, "f1": 0.2})
#     RunTable.create(data={"method": "random_forest", "precision": 2, "recall": 0.5, "f1_1": 0.6})
#     RunTable.merge_column("f1", "f1_1")
#     print(RunTable.as_dataframe())
